/*

Program : tcplib
Object  : manages TCP/IP connections
Author  : Jean-Marc DURO
EMail   : Jean-marc.duro@orange.fr
Status  :
  Opensource under LGPL expect for Telnet part (see below)
  Telnet part adapted from Franois PIETTE's TnCnx (see Copyright below)
Version : 2.11
Updates :
  Oct 23, 2015 V2.09 Ported Telnet part to OpenEuphoria 4
  Feb 18, 2016 V2.10 Separated Telnet part from TCP library
  Feb 19, 2016 V2.11 Added Server routines
Routines:
  function  circPrev(integer ptr)
  function  circNext(integer ptr)
  function  circDiff(integer read_ptr, integer write_ptr)
  function  circDataSize(integer cnx)
  function  circGetNewData(integer cnx)
  procedure circRazBuffer(integer cnx)
  function  errorLabel(integer err_num)
  function  tcpSend(integer cnx, sequence cmd)
  function  getActiveSockets()
  function  findConnection(sequence host, integer port)
  function  tcpReceive(integer cnx)
  function  tcpClose(integer cnx)
  procedure checkActiveConnections(integer timeout = 0, integer timeout_micro = 0)
  procedure checkPendingConnections(integer rtn_id=0, integer timeout = 0, integer timeout_micro = 0)
  function  tcpConnect(sequence host, integer port=80, integer protocol=NONE, integer rtn_id=0)
  function  tcpListen(sequence host, integer port=80, integer protocol=NONE)
  function tcpSendReceive(integer cnx, sequence cmd)
*/


/*

Program:    TNCNX.PAS
Object:     Delphi component which implement the TCP/IP tcp protocol
        including some options negociations.
        RFC854, RFC885, RFC779, RFC1091
Author:     Franois PIETTE
EMail:    http://users.swing.be/francois.piette  francois.piette@swing.be
        http://www.rtfm.be/fpiette       francois.piette@rtfm.be
        francois.piette@pophost.eunet.be
Creation:   April, 1996
Version:    2.09
Support:    Use the mailing list twsocket@rtfm.be See website for details.
Legal issues: Copyright (C) 1996-2002 by Franois PIETTE
        Rue de Grady 24, 4053 Embourg, Belgium. Fax: +32-4-365.74.56
        <francois.piette@pophost.eunet.be>

        This software is provided "as-is", without any express or
        implied warranty.  In no event will the author be held liable
        for any  damages arising from the use of this software.

        Permission is granted to anyone to use this software for any
        purpose, including commercial applications, and to alter it
        and redistribute it freely, subject to the following
        restrictions:

        1. The origin of this software must not be misrepresented,
         you must not claim that you wrote the original software.
         If you use this software in a product, an acknowledgment
         in the product documentation would be appreciated but is
         not required.

        2. Altered source versions must be plainly marked as such, and
         must not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any source
         distribution.

        4. You must register this software by sending a picture postcard
         to the author. Use a nice stamp and mention your name, street
         address, EMail address and any comment you like to say.

Updates:
Jul 22, 1997 Adapted to Delphi 3
Sep 5, 1997  Added version information, removed old code, added OnTermType
       Renamed some indentifier to be more standard.
Sep 24, 1997 V2.03 Added procedures to negociate options
May 12, 1998 V2.04 Changed telnetNegociateOption to properly handle unwanted
       option as Jan Tomasek <xtomasej@feld.cvut.cz> suggested.
Aug 10, 1998 V2.05 Cleared activeConnections[cnx][SUB_OPTION] after telnetNegociateSubOption as Jan
       Tomasek <xtomasej@feld.cvut.cz> suggested.
Aug 15, 1999 V2.06 Moved Notification procedure to public section for
       BCB4 compatibility
Aug 20, 1999 V2.07 Added compile time options. Revised for BCB4.
Jun 18, 2001 V2.08 Use AllocateHWnd and DeallocateHWnd from wsocket.
Oct 23, 2002 V2.09 Changed Buffer arg in OnDataAvailable to Pointer instead
       of PChar to avoid Delphi 7 messing everything with AnsiChar.

*/

include std/socket.e as sock
include std/text.e
include std/search.e
include std/sequence.e
include std/dll.e
--include std/os.e
include myLibs/myDebug.e
include myLibs/myConv.e
include myLibs/mySeq.e as seq
include myLibs/myFile.e
include myLibs/common.e

public constant
  -- Common
  TCP_HOST = 1, TCP_PORT = 2, TCP_SOCKET = 3, TCP_STATUS = 4,
  LAST_IO = 5, BUFFER = 6, WRITE_PTR = 7, READ_PTR = 8,
  PROTOCOL = 9, ROUTINE = 10,
  -- Telnet specific
  IAC = 11, VERB = 12, SUB_OPTION = 13, SUB_NEGOC = 14

public constant
  LSTN_PORT = 1, LSTN_SOCKET = 2, LSTN_STATUS = 3, LSTN_PROTOCOL = 4

-- TCP Status
public constant DISCONNECTED = 0, CONNECTED = 1

-- Protocol
public constant NONE = 0, TELNET = 1

public sequence activeConnections = {}, activeSockets = {}
public sequence listeningPorts = {}, listeningSockets = {}

object   void = 0

constant
  MAX_CHARS = 4096

public constant CRLF = "\r\n"

constant maxTime = 30

--------------------------------------------------------------------------------
-- Telnet part adapted from Franois PIETTE's TnCnx
--------------------------------------------------------------------------------

constant
  TnCnxVersion     = 209,
  CopyRight = " TTnCnx (c) 1996-2002 F. Piette V2.09 ",
  FLocation = "TNCNX",
  FTermType = "VT100",

  -- Telnet command characters                       
  TNCH_EOR        = 239,  -- #EF End Of Record (preceded by IAC)     
  TNCH_SE         = 240,  -- #F0 End of subnegociation parameters    
  TNCH_NOP        = 241,  -- #F1 No operation              
  TNCH_DATA_MARK  = 242,  -- #F2 Data stream portion of a Synch    
  TNCH_BREAK      = 243,  -- #F3 NVT charcater break           
  TNCH_IP         = 244,  -- #F4 Interrupt process           
  TNCH_AO         = 245,  -- #F5 Abort output              
  TNCH_AYT        = 246,  -- #F6 Are you there             
  TNCH_EC         = 247,  -- #F7 Erase character             
  TNCH_EL         = 248,  -- #F8 Erase line              
  TNCH_GA         = 249,  -- #F9 Go ahead                
  TNCH_SB         = 250,  -- #FA Subnegociation            
  TNCH_WILL       = 251,  -- #FB Will                  
  TNCH_WONT       = 252,  -- #FC Wont                  
  TNCH_DO         = 253,  -- #FD Do                  
  TNCH_DONT       = 254,  -- #FE Dont                  
  TNCH_IAC        = 255,  -- #FF IAC                   

  -- Telnet options                            
  TN_TRANSMIT_BINARY = 0,   -- #00 
  TN_ECHO            = 1,   -- #01 
  TN_RECONNECTION    = 2,   -- #02 
  TN_SUPPRESS_GA     = 3,   -- #03 
  TN_MSG_SZ_NEGOC    = 4,   -- #04 
  TN_STATUS          = 5,   -- #05 
  TN_TIMING_MARK     = 6,   -- #06 
  TN_NOPTIONS        = 6,   -- #06 
  TN_DET             = 20,  -- #14 
  TN_SEND_LOC        = 23,  -- #17 
  TN_TERMTYPE        = 24,  -- #18 
  TN_EOR             = 25,  -- #19 
  TN_NAWS            = 31,  -- #1F 
  TN_TERMSPEED       = 32,  -- #20 
  TN_TFC             = 33,  -- #21 
  TN_XDISPLOC        = 35,  -- #23 
  TN_EXOPL           = 255, -- #FF 

  TN_TTYPE_SEND      = 1,
  TN_TTYPE_IS        = 0

integer
  RemoteBinMode,
  LocalBinMode,
  FLocalEcho,
  spga,
  FTType

--------------------------------------------------------------------------------
-- End of Telnet part
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------

public function circPrev(integer ptr)
  ptr -= 1
  if ptr < 1 then ptr = MAX_CHARS end if
  return ptr
end function

--------------------------------------------------------------------------------

public function circNext(integer ptr)
  ptr += 1
  if ptr > MAX_CHARS then ptr = 1 end if
  return ptr
end function

--------------------------------------------------------------------------------

public function circDiff(integer read_ptr, integer write_ptr)
  if write_ptr >= read_ptr then
    return write_ptr - read_ptr
  else    
    return (MAX_CHARS - read_ptr) + write_ptr
  end if
end function

--------------------------------------------------------------------------------

procedure circAddChar(integer cnx, integer Ch)
  if circDiff(activeConnections[cnx][READ_PTR], activeConnections[cnx][WRITE_PTR]) = MAX_CHARS-1 then
    logMsg("circAddChar: Buffer overflow")
    return
  end if
  activeConnections[cnx][WRITE_PTR] = circNext(activeConnections[cnx][WRITE_PTR])
  activeConnections[cnx][BUFFER][ activeConnections[cnx][WRITE_PTR] ] = Ch
end procedure

--------------------------------------------------------------------------------

public function circDataSize(integer cnx)
  return circDiff(activeConnections[cnx][READ_PTR], activeConnections[cnx][WRITE_PTR])
end function

--------------------------------------------------------------------------------

public function circGetNewData(integer cnx)
  if activeConnections[cnx][WRITE_PTR] >= activeConnections[cnx][READ_PTR] then
    return activeConnections[cnx][BUFFER][ activeConnections[cnx][READ_PTR]+1..activeConnections[cnx][WRITE_PTR] ]
  else    
    return activeConnections[cnx][BUFFER][ activeConnections[cnx][READ_PTR]+1..MAX_CHARS ] &
            activeConnections[cnx][BUFFER][ 1..activeConnections[cnx][WRITE_PTR] ]
  end if
end function

--------------------------------------------------------------------------------

public procedure circRazBuffer(integer cnx)
  activeConnections[cnx][READ_PTR] = activeConnections[cnx][WRITE_PTR]
end procedure

--------------------------------------------------------------------------------

public function errorLabel(integer err_num)
  sequence lbl
  sequence status = sprintf("%d\t", err_num)
  switch err_num do
    case OK then
      lbl = "OK"
    case ERR_ACCESS then  -- -1
      lbl = "Permission has been denied"
    case ERR_ADDRINUSE then  -- -2
      lbl = "Address is already in use"
    case ERR_ADDRNOTAVAIL then  -- -3
      lbl = "The specified address is not a valid local IP address on this computer"
    case ERR_AFNOSUPPORT then  -- -4
      lbl = "Address family not supported by the protocol family"
    case ERR_AGAIN then  -- -5
      lbl = "Kernel resources to complete the request are temporarly unavailable"
    case ERR_ALREADY then  -- -6
      lbl = "Operation is already in progress"
    case ERR_CONNABORTED then  -- -7
      lbl = "Software has caused a connection to be aborted"
    case ERR_CONNREFUSED then  -- -8
      lbl = "Connection was refused"
    case ERR_CONNRESET then  -- -9
      lbl = "An incoming connection was supplied however it was terminated by the remote peer"
    case ERR_DESTADDRREQ then  -- -10
      lbl = "Destination address required"
    case ERR_FAULT then  -- -11
      lbl = "Address creation has failed internally"
    case ERR_HOSTUNREACH then  -- -12
      lbl = "No route to the host specified could be found"
    case ERR_INPROGRESS then  -- -13
      lbl = "A blocking call is inprogress"
    case ERR_INTR then  -- -14
      lbl = "A blocking call was cancelled or interrupted"
    case ERR_INVAL then  -- -15
      lbl = "An invalid sequence of command calls were made"
    case ERR_IO then  -- -16
      lbl = "An I/O error occurred while making the directory entry or allocating the inode"
    case ERR_ISCONN then  -- -17
      lbl = "Socket is already connected"
    case ERR_ISDIR then  -- -18
      lbl = "An empty pathname was specified"
    case ERR_LOOP then  -- -19
      lbl = "Too many symbolic links were encountered"
    case ERR_MFILE then  -- -20
      lbl = "The queue is not empty upon routine call"
    case ERR_MSGSIZE then  -- -21
      lbl = "Message is too long for buffer size"
    case ERR_NAMETOOLONG then  -- -22
      lbl = "Component of the path name exceeded 255 characters or the entire path exceeded 1023 characters"
    case ERR_NETDOWN then  -- -23
      lbl = "The network subsystem is down or has failed"
    case ERR_NETRESET then  -- -24
      lbl = "Network has dropped it's connection on reset"
    case ERR_NETUNREACH then  -- -25
      lbl = "Network is unreachable"
    case ERR_NFILE then  -- -26
      lbl = "Not a file"
    case ERR_NOBUFS then  -- -27
      lbl = "No buffer space is available"
    case ERR_NOENT then  -- -28
      lbl = "Named socket does not exist"
    case ERR_NOTCONN then  -- -29
      lbl = "Socket is not connected"
    case ERR_NOTDIR then  -- -30
      lbl = "Component of the path prefix is not a directory"
    case ERR_NOTINITIALISED then  -- -31
      lbl = "Socket system is not initialized"
    case ERR_NOTSOCK then  -- -32
      lbl = "The descriptor is not a socket"
    case ERR_OPNOTSUPP then  -- -33
      lbl = "Operation is not supported on this type of socket"
    case ERR_PROTONOSUPPORT then  -- -34
      lbl = "Protocol not supported"
    case ERR_PROTOTYPE then  -- -35
      lbl = "Protocol is the wrong type for the socket"
    case ERR_ROFS then  -- -36
      lbl = "The name would reside on a read-only file system"
    case ERR_SHUTDOWN then  -- -37
      lbl = "The socket has been shutdown"
    case ERR_SOCKTNOSUPPORT then  -- -38
      lbl = "Socket type is not supported"
    case ERR_TIMEDOUT then  -- -39
      lbl = "Connection has timed out"
    case ERR_WOULDBLOCK then  -- -40
      lbl = "The operation would block on a socket marked as non-blocking"
    case else
      lbl = "Unknown error!"
  end switch
  return status&lbl
end function

--------------------------------------------------------------------------------

public function tcpSend(integer cnx, sequence cmd)
  if length(activeConnections)=0 then return 0 end if
  if length(cmd)=0 then return 0 end if
  if activeConnections[cnx][TCP_STATUS] = DISCONNECTED then return 0 end if
  sequence host = activeConnections[cnx][TCP_HOST]
  sock:socket tcpSocket = activeConnections[cnx][TCP_SOCKET]
  integer protocol = activeConnections[cnx][PROTOCOL]
  if (protocol=TELNET) and ends(cmd, "\n") and not ends(cmd, "\r\n") then
    cmd = cmd[1..$-1] & "\r\n"
  end if
  integer status = sock:send(tcpSocket,cmd,0)
  if status>=0 then
    logMsg(sprintf("tcpSend: Sent '%s' %s (%d bytes) to host %s", {showPrintable(cmd), sprint(cmd), status, host}))
    activeConnections[cnx][LAST_IO] = time()
  else
    logMsg(sprintf("tcpSend: Error '%s' while sending '%s' to host %s", {errorLabel(status), cmd, host}))
    if (status = ERR_CONNRESET) then
      logMsg(sprintf("tcpSend: lost connection to host %s", {host}))
      activeConnections[cnx][TCP_STATUS] = DISCONNECTED
    end if
  end if
  return status
end function

--------------------------------------------------------------------------------

public function getActiveSockets()
  sequence s = {}
  for i = 1 to length(activeConnections) do
    if activeConnections[i][TCP_STATUS] = CONNECTED then
      s = append(s, activeConnections[i][TCP_SOCKET])
    end if
  end for
  return s
end function

--------------------------------------------------------------------------------

public function getListeningSockets()
  sequence s = {}
  for i = 1 to length(listeningPorts) do
    if listeningPorts[i][LSTN_STATUS] = CONNECTED then
      s = append(s, listeningPorts[i][LSTN_SOCKET])
    end if
  end for
  return s
end function

--------------------------------------------------------------------------------

public function findConnection(sequence host, integer port)
  sequence s = commonFields(activeConnections, {TCP_HOST,TCP_PORT}, {host,port})
  if length(s) then return s[1] else return 0 end if
end function

--------------------------------------------------------------------------------

public function findPort(integer port)
  sequence s = commonFields(listeningPorts, {LSTN_PORT}, {port})
  if length(s) then return s[1] else return 0 end if
end function

--------------------------------------------------------------------------------
-- Telnet part adapted from Franois PIETTE's TnCnx
--------------------------------------------------------------------------------

procedure telnetAnswer(integer cnx, integer chAns, integer chOption)
  sequence Buf = {TNCH_IAC, chAns, chOption}
  tcpSend(cnx, Buf)
end procedure

--------------------------------------------------------------------------------

procedure telnetNegociateSubOption(integer cnx, sequence strSubOption)
  switch strSubOption[1] do
  case TN_TERMTYPE then
    if strSubOption[2] = TN_TTYPE_SEND then
      sequence Buf = {TNCH_IAC, TNCH_SB, TN_TERMTYPE, TN_TTYPE_IS} & FTermType & {TNCH_IAC, TNCH_SE}
      tcpSend(cnx, Buf)
    end if
  case else
--    puts(f_debug, "Unknown suboption\r\n"); 
  end switch
end procedure

--------------------------------------------------------------------------------

procedure telnetNegociateOption(integer cnx, integer chAction, integer chOption)
  switch chOption do
  case TN_TRANSMIT_BINARY then
      if chAction = TNCH_WILL then
        telnetAnswer(cnx, TNCH_DO, chOption)
        RemoteBinMode = 1
        LocalBinMode  = 1
      elsif chAction = TNCH_WONT then
        if RemoteBinMode then
          RemoteBinMode = 0
          LocalBinMode  = 0
        end if
      end if
  case TN_ECHO then
      if chAction = TNCH_WILL then
        telnetAnswer(cnx, TNCH_DO, chOption)
        FLocalEcho = 0
      elsif chAction = TNCH_WONT then
        FLocalEcho = 1
      end if
  case TN_SUPPRESS_GA then
      if chAction = TNCH_WILL then
        telnetAnswer(cnx, TNCH_DO, chOption)
        spga = 1
      end if
  case TN_TERMTYPE then
      if chAction = TNCH_DO then
        telnetAnswer(cnx, TNCH_WILL, chOption)
        FTType = 1
      end if
  case TN_SEND_LOC then
      if chAction = TNCH_DO then
        telnetAnswer(cnx, TNCH_WILL, chOption)
        sequence Buf = {TNCH_IAC, TNCH_SB, TN_SEND_LOC, FLocation, TNCH_IAC, TNCH_SE}
        tcpSend(cnx, Buf)
      end if
  case TN_EOR then
      if chAction = TNCH_DO then
        telnetAnswer(cnx, TNCH_WILL, chOption)
        FTType = 1
      end if
  case else
--    telnetAnswer(cnx, TNCH_WONT, chOption); 
    -- Jan Tomasek <xtomasej@feld.cvut.cz> 
    if chAction = TNCH_WILL then
      telnetAnswer(cnx, TNCH_DONT, chOption)
    else
      telnetAnswer(cnx, TNCH_WONT, chOption)
    end if
  end switch
end procedure

--------------------------------------------------------------------------------

procedure telnetReceiveChar(integer cnx, integer Ch)

  if activeConnections[cnx][VERB] != #0 then
    telnetNegociateOption(cnx, activeConnections[cnx][VERB], Ch)
    activeConnections[cnx][VERB]     = #0
    activeConnections[cnx][SUB_OPTION] = ""
    return
  end if

  if activeConnections[cnx][SUB_NEGOC] then
    if Ch = TNCH_SE then
      activeConnections[cnx][SUB_NEGOC]  = 0
      telnetNegociateSubOption(cnx, activeConnections[cnx][SUB_OPTION])
      activeConnections[cnx][SUB_OPTION] = ""
    else
      activeConnections[cnx][SUB_OPTION] = activeConnections[cnx][SUB_OPTION] & Ch
    end if
    return
  end if

  if activeConnections[cnx][IAC] then
    switch Ch do
    case TNCH_IAC then
            circAddChar(cnx, Ch)
            activeConnections[cnx][IAC] = 0
    case TNCH_DO, TNCH_WILL, TNCH_DONT, TNCH_WONT then
            activeConnections[cnx][IAC]   = 0
            activeConnections[cnx][VERB] = Ch
    case TNCH_EOR then
        puts(f_debug, "TNCH_EOR\r\n")
        activeConnections[cnx][IAC]   = 0
    case TNCH_SB then
--        puts(f_debug, "Subnegociation\r\n"); 
        activeConnections[cnx][SUB_NEGOC] = 1
        activeConnections[cnx][IAC]    = 0
    case else
      printf(f_debug, "Unknown %02x\r\n", Ch)
      activeConnections[cnx][IAC] = 0
    end switch
    return
  end if

  switch Ch do
  case TNCH_EL then
      puts(f_debug, "TNCH_EL\r\n")
      circAddChar(cnx, Ch)
  case TNCH_EC then
      puts(f_debug, "TNCH_EC\r\n")
      circAddChar(cnx, Ch)
  case TNCH_AYT then
      puts(f_debug, "TNCH_AYT\r\n")
      circAddChar(cnx, Ch)
  case TNCH_IP then
      puts(f_debug, "TNCH_IP\r\n")
      circAddChar(cnx, Ch)
  case TNCH_AO then
      puts(f_debug, "TNCH_AO\r\n")
      circAddChar(cnx, Ch)
  case TNCH_IAC then
      activeConnections[cnx][IAC] = 1
  case else
    circAddChar(cnx, Ch)
  end switch
end procedure

--------------------------------------------------------------------------------
-- End of Telnet part
--------------------------------------------------------------------------------

public function tcpReceive(integer cnx)
  if length(activeConnections)=0 then return 0 end if
  if activeConnections[cnx][TCP_STATUS] = DISCONNECTED then return 0 end if
  sequence host = activeConnections[cnx][TCP_HOST]
  sock:socket tcpSocket = activeConnections[cnx][TCP_SOCKET]
  integer protocol = activeConnections[cnx][PROTOCOL]
  object data = sock:receive(tcpSocket, 0)
  if atom(data) then
    if find(data,{ERR_CONNABORTED,ERR_CONNREFUSED,ERR_CONNRESET,ERR_NETDOWN,ERR_NETRESET,ERR_NETUNREACH}) then
      logMsg(sprintf("tcpReceive: lost connection from host %s", {host}))
      activeConnections[cnx][TCP_STATUS] = DISCONNECTED
    else
      if (data != OK) then
        logMsg(sprintf("tcpReceive: Error %s while receiving from host %s", {errorLabel(data),host}))
      end if
    end if
  else
    activeConnections[cnx][LAST_IO] = time()
    if length(data) then
      switch protocol do
      case TELNET then
        for i = 1 to length(data) do telnetReceiveChar(cnx, data[i]) end for
      case else
        for i = 1 to length(data) do circAddChar(cnx, data[i]) end for
      end switch
      data = circGetNewData(cnx)
      analyzeObject(data, sprintf("tcpReceive(%d)", {cnx}), f_debug)
     end if
  end if
  circRazBuffer(cnx)
  return data
end function

--------------------------------------------------------------------------------

public function tcpClose(integer cnx)
  if length(activeConnections)=0 then return 0 end if
  if activeConnections[cnx][TCP_STATUS] = DISCONNECTED then return 0 end if
  sequence host = activeConnections[cnx][TCP_HOST]
  sock:socket tcpSocket = activeConnections[cnx][TCP_SOCKET]
  integer status = sock:close(tcpSocket)
  if status != 0 then
    logMsg(sprintf("tcpClose: error %s closing connection to host %s", {errorLabel(status), host}))
  else
    --activeConnections = remove(activeConnections, cnx)
    activeConnections[cnx][TCP_STATUS] = DISCONNECTED
    activeSockets = getActiveSockets()
  end if
  return status
end function

--------------------------------------------------------------------------------

public procedure checkActiveConnections(integer timeout = 0, integer timeout_micro = 0)
  -- get socket state
  object sock_data = sock:select(activeSockets, {}, activeSockets,
                          timeout, timeout_micro)
  if atom(sock_data) then
    logMsg(sprintf("checkActiveConnections: Error %s", {errorLabel(sock_data)}))
    return
  end if
  for i = 1 to length(sock_data) do
    integer cnx = findRecNumByField(activeConnections, TCP_SOCKET, sock_data[i][SELECT_SOCKET])
    if cnx then
      sequence host = activeConnections[cnx][TCP_HOST]
      integer port = activeConnections[cnx][TCP_PORT]
      integer rtn_id = activeConnections[cnx][ROUTINE]
      if sock_data[i][SELECT_IS_READABLE] then
        --logMsg(sprintf("checkConnections: Data readable from host %s", {host}))
        object data = tcpReceive(cnx)
        if sequence (data) then
          if length(data) and rtn_id then  -- process data with routine rtn_id
            call_proc(rtn_id, {cnx, data})
          end if
        elsif (data != OK) then
          printf(f_debug, "checkActiveConnections: data = %d, close Telnet\n", data)
          void = tcpClose(cnx)
        end if
      elsif (time() - activeConnections[cnx][LAST_IO]) >= maxTime then
        void = tcpSend(cnx, {32,8})  -- SPACE and BACKSPACE
        activeConnections[cnx][LAST_IO] = time()
      end if
      if sock_data[i][SELECT_IS_ERROR] then
        logMsg(sprintf("checkActiveConnections: Connexion to host %s is in error state", {host}))
      end if
    end if
  end for
end procedure

--------------------------------------------------------------------------------

public procedure checkPendingConnections(integer rtn_id=0, integer timeout = 0, integer timeout_micro = 0)
  -- get socket state
  object sock_data = sock:select(listeningSockets, {}, listeningSockets,
                          timeout, timeout_micro)
  if atom(sock_data) then
    logMsg(sprintf("checkPendingConnections: Error %s", {errorLabel(sock_data)}))
    return
  end if
  for i = 1 to length(sock_data) do
    integer cnx = findRecNumByField(listeningPorts, LSTN_SOCKET, sock_data[i][SELECT_SOCKET])
    if cnx then
      integer port = listeningPorts[cnx][LSTN_PORT]
      sock:socket tcpSocket = listeningPorts[cnx][LSTN_SOCKET]
      integer protocol = listeningPorts[cnx][LSTN_PROTOCOL]
      if sequence(sock_data[i][SELECT_IS_READABLE]) then
        logMsg(sprintf("checkPendingConnections: Data readable from port %s", {port}))
        object client = sock:accept(tcpSocket)  --{socket client, sequence client_ip_address}
        if atom(client) then
          logMsg(sprintf("checkPendingConnections: Could not accept connections on port %d, error %s", {port,errorLabel(client)}))
          void = tcpClose(cnx)
          return
        end if
        logMsg(sprintf("checkPendingConnections: Client %s connected via socket %s", {client[2], sprint(client[1])}))
        switch protocol do
          case TELNET then
            activeConnections = append(activeConnections, {client[2], port, client[1], CONNECTED, time(), repeat(0, MAX_CHARS), 0, 0, TELNET, rtn_id, 0, #0, "", 0})
          case else
            activeConnections = append(activeConnections, {client[2], port, client[1], CONNECTED, time(), repeat(0, MAX_CHARS), 0, 0, NONE, rtn_id})
        end switch
        activeSockets = getActiveSockets()
      end if
      if sock_data[i][SELECT_IS_ERROR] then
        logMsg(sprintf("checkActiveConnections: Connexion to port %d is in error state", {port}))
      end if
    end if
  end for
end procedure

--------------------------------------------------------------------------------

public function tcpConnect(sequence host, integer port=80, integer protocol=NONE, integer rtn_id=0)
  sock:socket tcpSocket = sock:create(AF_INET,SOCK_STREAM,0)
  if atom(tcpSocket) then
    logMsg("tcpConnect: Could not create socket, error "&errorLabel(tcpSocket))
    return 0
  end if
  --logMsg("tcpConnect: Socket "&sprint(tcpSocket)&" created")
  integer status = sock:connect(tcpSocket, host, port)
  if status < 0 then
    logMsg(sprintf("tcpConnect: Could not connect to %s:%d, error %s", {host, port, errorLabel(status)}))
    void = sock:close(tcpSocket)
    return 0
  end if
  logMsg(sprintf("tcpConnect: Connected to %s:%d", {host, port}))
  switch protocol do
    case TELNET then
      activeConnections = append(activeConnections, {host, port, tcpSocket, CONNECTED, time(), repeat(0, MAX_CHARS), 0, 0, TELNET, rtn_id, 0, #0, "", 0})
    case else
      activeConnections = append(activeConnections, {host, port, tcpSocket, CONNECTED, time(), repeat(0, MAX_CHARS), 0, 0, NONE, rtn_id})
  end switch
  activeSockets = getActiveSockets()
  return length(activeConnections)
end function

--------------------------------------------------------------------------------

public function tcpListen(sequence host, integer port=80, integer protocol=NONE)
  sock:socket tcpSocket = sock:create(AF_INET,SOCK_STREAM,0)
  if atom(tcpSocket) then
    logMsg("tcpListen: Could not create socket, error "&errorLabel(tcpSocket))
    return 0
  end if
  --logMsg("tcpListen: Socket "&sprint(tcpSocket)&" created")
  sequence addr = sprintf("0.0.0.0:%d",port)
  integer status = sock:bind(tcpSocket, addr)
  if status != sock:OK then
    logMsg(sprintf("tcpListen: Binding to %s has failed, error %s", 
               { addr, errorLabel(sock:error_code()) }))
    void = sock:close(tcpSocket)
    return 0
  end if
  --logMsg("tcpListen: Socket "&sprint(tcpSocket)&" bound to "&addr)
  status = sock:listen(tcpSocket, 0)
  if status != sock:OK then
    logMsg(sprintf("tcpListen: Listening on socket %s has failed, error %s", 
               { sprint(tcpSocket), errorLabel(sock:error_code()) }))
    void = sock:close(tcpSocket)
    return 0
  end if
  logMsg(sprintf("tcpListen: Listening on socket %s", { sprint(tcpSocket) }))
  listeningPorts = append(listeningPorts, {port, tcpSocket, CONNECTED, protocol})
  listeningSockets = getListeningSockets()
  return length(listeningPorts)
end function

------------------------------------------------------------------------------
-- COMPLEX FUNCTIONS
------------------------------------------------------------------------------

public function tcpSendReceive(integer cnx, sequence cmd)
  object status = tcpSend(cnx, cmd&CRLF)
  if status<0 then
    logMsg("... called from tcpSendReceive")
    return -1
  end if
  object data = tcpReceive(cnx)
  if atom(data) then
    logMsg("... called from tcpSendReceive")
    return -1
  else
    return data
  end if
end function

