{ 
    This program is free software; you can redistribute it and/or modify 
    it under the terms of the GNU General Public License as published by 
    the Free Software Foundation; either version 3 of the License, or 
    (at your option) any later version. 
 
    This program is distributed in the hope that it will be useful, 
    but WITHOUT ANY WARRANTY; without even the implied warranty of 
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
    GNU General Public License for more details. 
 
    You should have received a copy of the GNU General Public License 
    along with this program.  If not, see . 
} 
unit Common; 
 
{$mode objfpc}{$H+}

interface

uses Classes, SysUtils, Dialogs, StdCtrls, Grids, FileUtil, Process, LazFileUtils;

const
  CrLf = #13+#10;
  Tab = #9;

type
  {An exception which is raised in TStringGrid when trying to load data
  from an invalid file.}
  EInvalidFile = class(Exception);

  {This exception is raised in TStringGrid when trying to read strings
  from file, and there are more strings in file than fits into the grid.}
  ETooSmallGrid = class(EXception);

  TDesc = record
    FId  : integer;
    FName: string;
  end;
  PDesc = ^TDesc;

  TDescList = class(TList) { a customized version of TList to hold PDesc pointers }
  private
    procedure  SetDesc(index: Integer; Ptr: PDesc);
    function   GetDesc(index: Integer): PDesc;
  public
    procedure  Clear; override;
    destructor Destroy; override;
    function   AddDesc(num: integer; desc: string): integer;
    property   Descriptions[i: Integer]: PDesc read GetDesc write SetDesc;
  end;

  TStringArray = array of string;
  TStringMatrix  = array of TStringArray;

  TDirection = (Normal, Reverse);

var
  f_debug: TextFile;
  InitialDir: string;

function  dateStamp(msg: string): string;
procedure logMsg(msg: string);
function  mid(s: string; st, en: integer): string;
procedure warnErr(const Msg: String);
function  bytes2uint(a: PByteArray; len: byte; direction: TDirection=Reverse): cardinal;
function  splitString(str: string; sep: char; var l: TStringList): integer;
function  joinStrings(Elements: TStringList; separateur: AnsiChar): string;
function  rightPad(chaine: string; longueur: byte): string;
function  leftPad(chaine: string; longueur: byte): string;
function  NumStr(n: Word; d: byte): string;
function  StrToNum(s : string; b: byte) : longint;
function  Value(s: string): extended;
function  findRecord(List: TStringList; s: string): integer;
function  PosEx(const SubStr, S: string; Offset: Cardinal): integer;
function  Quote(s: string): string;
function  unQuote(s: string): string;
procedure execCommand(cmd: String; params: TStrings; var AStringList: TStringList);
procedure QuickSort(var A: array of Integer; iLo, iHi: Integer) ;
function  cleanStr(const S: string): string;
function  escapeStr(aString: string): string;
function  replaceString(str, sub1, sub2: string): string;
procedure copyFile(old_file, new_file: string);
procedure textReplace(old_file, new_file, old, new: string);
procedure gridClean(var grid: TStringGrid; fields: TStringList);
procedure gridLoadFromFile(var grid: TStringGrid; filename: string);
procedure gridSaveToFile(var grid: TStringGrid; filename: string);
procedure gridExchangeRows(var grid: TStringGrid; i, j: Integer );
procedure gridSort(var grid: TStringGrid; byColumn: Integer; iLo, iHi: Integer; CaseSensitive: Boolean);
procedure gridInvert(var grid: TStringGrid);
procedure memoLoad(var memo: TMemo; fichier: string);
procedure memoSave(memo: TMemo; fichier: string);
procedure listboxLoad(var liste: TListBox; fichier: string);
procedure listboxSave(liste: TListBox; fichier: string);
function  TryStrToInt(const S: string; out Value: Integer): Boolean;
function  AnsiReplaceText(str, sub1, sub2: String): String;
function  IncludeTrailingPathDelimiter(const S: string): string;
function  HexString(const t: PByteArray; size: Integer): string;
function  HexDump(const t: PByteArray; size: Integer): string;
function  readCSV(filename: string; var MyArray : TStringMatrix; var fields: TStringList): longint;
procedure saveCSV(filename: string; var MyArray : TStringMatrix; var fields: TStringList);
procedure readDescription(FileName: string; var descList: TDescList);
function  findDescription(descList: TDescList; id: integer): string;

implementation

//----------------------------------------------------------------------------

{Methods of TDescList}

//----------------------------------------------------------------------------

procedure TDescList.SetDesc(index: Integer; Ptr: PDesc);
var
  p: PDesc;
begin
  { get the pointer currently in slot index }
  p := Descriptions[index];
  if p <> Ptr then begin
    { if it is different from the one we are asked to put into this slot, check if it is <> Nil. If so, dispose of the memory it points at! }
    if p <> nil then Dispose(p);
    { store the passed pointer into the slot }
    Items[index] := Ptr;
  end;
end;

//----------------------------------------------------------------------------

function TDescList.GetDesc(index: Integer): PDesc;
begin
  { return the pointer in slot index, typecast to PDesc }
  Result := PDesc(Items[index]);
end;

//----------------------------------------------------------------------------

procedure TDescList.Clear;
var
  i: Integer;
  p: PDesc;
begin
  { dispose of the memory pointed to by all pointers in the list that are not Nil }
  for i := 0 to Pred(Count) do begin
    p := Descriptions[i];
    if p <> nil then Dispose(p);
  end;
  { call the Clear method inherited from TList to set Count to 0 }
  inherited Clear;
end;

//----------------------------------------------------------------------------

destructor TDescList.Destroy;
begin
  { clear the list to dispose of any pointers still stored first }
  Clear;
  inherited Destroy;
end;

//----------------------------------------------------------------------------

function TDescList.AddDesc(num: integer; desc: string): integer;
var
  Ptr: PDesc; { local variable in a method }
begin
  New(Ptr); { allocate a record on the heap }
  with Ptr^ do begin { note the caret to dereference the pointer }
    { put data into the fields of the record }
    FId := num;
    FName := desc;
  end;
  Result := Add(Ptr);
end;

//----------------------------------------------------------------------------

function dateStamp(msg: string): string;
// prefixes the message by date and time for logging
begin
  Result := DateTimeToStr(now) + ' -> ' + msg;
end;

//----------------------------------------------------------------------------

procedure logMsg(msg: string);
// records logs
begin
  try
    writeLn(f_debug, dateStamp(msg));
    flush(f_debug);
  except
    on E: EInOutError do
     writeln('File handling error occurred. Details: ', E.Message);
  end;
end;

//-----------------------------------------------------------------------------

function mid(s: string; st, en: integer): string;
begin
  Result := copy(s, st, en-st+1);
end;

//----------------------------------------------------------------------------

procedure warnErr(const Msg: String);
begin
  //messageDlg(Msg,mtError,[mbOk],0);
  WriteLn(Msg);
end;

//----------------------------------------------------------------------------

// converts an integer to a sequence of bytes
// digits is the length of the resulting sequence
// direction is Normal or Reverse (default)
// when direction is Reverse (usual), lowest byte is first
// when direction is Normal (telecom frames), highest byte is first
// example:
//   uint2bytes($1C0A0000, 4, Normal)  = ($1C, $0A, $00, $00)
//   uint2bytes($1C0A0000, 4, Reverse) = ($00, $00, $0A, $1C)
function bytes2uint(a: PByteArray; len: byte; direction: TDirection=Reverse): cardinal;
var
  i: byte;
  l, mult: longint;
begin
  l := 0;
  mult := 1;
  case Direction of
    Normal:
      begin
        for i := len-1 downto 0 do begin
          l := l + a^[i]*mult;
          mult := mult * $100;
        end;
      end;
    Reverse:
      begin
        for i := 0 to len-1 do begin
          l := l + a^[i]*mult;
          mult := mult * $100;
        end;
      end;
  end;
  Result := l;
end;

//----------------------------------------------------------------------------

function splitString(str: string; sep: char; var l: TStringList): integer;
// splits string str in tokens according to delimiter sep
var
  i: integer;
  s: string;
  inString: boolean;
begin
  l.Clear;
  s := '';
  inString := false;      // si 1, on est entre deux guillemets
  for i := 1 to length(str) do begin
    if (str[i] = '"') and (sep <> '"') then
      inString := not inString
    else
      if (not inString) and (str[i] = sep) then begin
        l.Add(s);
        s := '';
      end else
	s := s + str[i];
  end;
  l.Add(s);
  Result := l.Count;
end;

//----------------------------------------------------------------------------

function joinStrings(Elements: TStringList; separateur: AnsiChar): string;
var
  s: string;
  i: byte;
begin
  if Elements.Count = 0 then begin
    Result := '';
    exit;
  end;
  s := Elements[0];
  for i := 1 to Elements.Count do s := s + separateur + Elements[i];
  Result := s;
end;

//----------------------------------------------------------------------------

function rightPad(chaine: string; longueur: byte): string;
var
  lg: byte;
begin
  lg := length(chaine);
  if lg >= longueur then Result := chaine;
  Result := chaine + StringOfChar(' ', longueur-lg);
end;

//----------------------------------------------------------------------------

function leftPad(chaine: string; longueur: byte): string;
var
  lg: byte;
begin
  lg := length(chaine);
  if lg >= longueur then Result := chaine;
  Result := StringOfChar(' ', longueur-lg) + chaine;
end;

//----------------------------------------------------------------------------

function NumStr(n: Word; d: byte): string;
begin
  SetLength(Result, d);
  while d>0 do begin
    Result[d] := chr(n mod 10 + ord('0'));
    n := n div 10;
    dec(D);
  end;
end;

//----------------------------------------------------------------------------

function StrToNum(s : string; b: byte) : longint;
var
  n, i: longint;
  d: byte;
begin
  n := 0;
  d := Length(s);
  i := 1;
  while d>0 do begin
    case s[d] of
      '0'..'9':
         n := n + (Ord(s[d])-48)*i;
      'A'..'F':
         n := n + (Ord(s[d])-55)*i;
      'a'..'f':
         n := n + (Ord(s[d])-87)*i;
    end;
    Dec(d);
    i := i * b;
  end;
  StrToNum := n;
end;

//----------------------------------------------------------------------------

function Value(s: string): extended;
var
  p: byte;
begin
  p := pos('.',s);
  if p > 0 then s[p] := ',';
  Result := StrToFloat(s);
end;

//-----------------------------------------------------------------------------

function findRecord(List: TStringList; s: string): integer;
var
  i: integer;
begin
  Result := -1;
  for i := 1 to List.Count do
    if pos(LowerCase(s), LowerCase(List[i-1])) > 0 then Result := i-1;
end;

//-----------------------------------------------------------------------------

function PosEx(const SubStr, S: string; Offset: Cardinal): integer;
begin
  PosEx := 0;
  if (length(SubStr) > 0) and (Offset > 0) and (Offset <= Cardinal(length(S))) then
    PosEx := pos(SubStr, mid(S, Offset, length(S))) + Integer(Offset) - 1;
end;

//-----------------------------------------------------------------------------

function Quote(s: string): string;
begin
  if length(s) = 0 then
    Result := '""'
  else if s[1] <> '"' then
    Result := '"'+s+'"'
  else
    Result := s;
end;

//-----------------------------------------------------------------------------

function unQuote(s: string): string;
var
  lg: integer;
begin
  lg := length(s);
  if lg < 2 then
    Result := s
  else if (s[1] = '"') and (s[lg] = '"') then
    Result := copy(s, 2, lg-2)
  else
    Result := s;
end;

//----------------------------------------------------------------------------

procedure execCommand(cmd: String; params: TStrings; var AStringList: TStringList);
var
  AProcess: TProcess;
begin
  //WriteLn(cmd);
  // Now we will create the TProcess object, and
  // assign it to the var AProcess.
  AProcess := TProcess.Create(nil);

  // Tell the new AProcess what the command to execute is.
  // Let's use the Free Pascal compiler

  AProcess.Executable := cmd;
  AProcess.Parameters := params;

  // Previous versions of FPC did not yet have .Executable and .Parameters and
  // used the .CommandLine property. While this is still available, it is deprecated
  // so it should not be used.


  // We will define an option for when the program
  // is run. This option will make sure that our program
  // does not continue until the program we will launch
  // has stopped running. Also now we will tell it that
  // we want to read the output of the file.
  AProcess.Options := AProcess.Options + [poWaitOnExit, poUsePipes];

  // Now that AProcess knows what the commandline is
  // we will run it.
  AProcess.Execute;

  // This is not reached until ppc386 stops running.

  // Now read the output of the program we just ran
  // into the TStringList.
  AStringList.LoadFromStream(AProcess.Output);

  // Now that the file is saved we can free the TProcess.
  AProcess.Free;
end;

//----------------------------------------------------------------------------

procedure QuickSort(var A: array of Integer; iLo, iHi: Integer) ;
var
  Lo, Hi, Pivot, T: Integer;
begin
  Lo := iLo;
  Hi := iHi;
  Pivot := A[(Lo + Hi) div 2];
  repeat
    while A[Lo] < Pivot do Inc(Lo) ;
    while A[Hi] > Pivot do Dec(Hi) ;
    if Lo <= Hi then begin
      T := A[Lo];
      A[Lo] := A[Hi];
      A[Hi] := T;
      Inc(Lo) ;
      Dec(Hi) ;
    end;
  until Lo > Hi;
  if Hi > iLo then QuickSort(A, iLo, Hi) ;
  if Lo < iHi then QuickSort(A, Lo, iHi) ;
end;

//----------------------------------------------------------------------------

// convert a string to escape string format as in c syntax
// when sending a query with blobs a blob should be escaped and enclosed between '' or "" 
 
function escapeStr(aString: string): string; 
var 
  I, j, k: Integer; 
  pc: PChar; 
begin 
  j := 0; 
  k := Length(aString); 
  //how many chars needs escape mark? 
  for I := 1 to k do
    if aString[I] in ['''', '"', '\', #9, #10, #13, #0] then Inc(j); 
  SetLength(Result, k + j); //dimension the result to right size 
  pc := PChar(Result); //a pchar for walking the result 
  for I := 1 to k do begin 
    if aString[I] in ['''', '"', '\', #9, #10, #13, #0] then begin //if this char needs escape 
      pc^ := '\'; //escape 
      Inc(pc); //next char 
      case aString[I] of 
        #9: pc^ := 't'; 
        #10: pc^ := 'n'; 
        #13: pc^ := 'r'; 
        #0: pc^ := '0'; 
      else 
        pc^ := aString[I]; 
      end; 
    end 
    else 
      pc^ := aString[I]; 
    Inc(pc); //next char 
  end; 
end; 
 
//----------------------------------------------------------------------------

// remove unreadable chars from string

function cleanStr(const S: string): string; 
const 
  Remove = [#00, #09, #32, #13, #10]; 
var 
  i: integer; 
  doubleSpace: boolean; 
begin 
  result := ''; 
  doubleSpace := true; 
  for i := 1 to length(S) do begin 
    if (S[i] in remove) then begin 
      if not doubleSpace then result := result + #32; 
      doubleSpace := true; 
    end else begin 
      result := result + S[i]; 
      doubleSpace := false; 
    end; 
  end;
end; 
 
//----------------------------------------------------------------------------

procedure gridClean(var grid: TStringGrid; fields: TStringList);
var
  i, cols: integer;
begin
  if fields.Count > grid.ColCount then cols := grid.ColCount else cols := fields.Count;
  for i := 0 to cols-1 do grid.Cols[i].Clear;
  grid.RowCount := 2;
  grid.FixedRows := 1;
  for i := 0 to cols-1 do grid.Cells[i,0] := fields[i];
end;

//----------------------------------------------------------------------------

procedure gridLoadFromFile(var grid: TStringGrid; filename: string);
var
  f_in:TextFile;
  r:LongInt;
  c:LongInt;
  Inn:string;
  Elements: TStringList;
  n: byte;
begin
  try
    Assignfile(f_in,filename);
    Reset(f_in);
    Elements := TStringList.Create;
    if eof(f_in) then
      raise EInvalidFile.Create('Invalid file. Program cannot read data');
    r := 0;
    with grid do begin
      repeat
        if r > rowcount-1 then
          rowcount := rowcount+1;
        Readln(f_in,inn);
        n := splitString(inn, ';', Elements);
        if n > colcount then
          raise ETooSmallGrid.Create('Grid too small. Not enough columns');
        for c := 0 to n do
          cells[c,r] := Elements[c];
        Inc(r);
      until eof(f_in);
    end;
    if not eof(f_in) then
      raise ETooSmallGrid.Create('Grid too small. Could not load all data');
  Finally
    Closefile(f_in);
    Elements.Free;
  end;
end;

//----------------------------------------------------------------------------

procedure gridSaveToFile(var grid: TStringGrid; filename: string);
var
  f_out: TextFile;
  r:LongInt;
  c:LongInt;
begin
  try
    Assignfile(f_out,filename);
    Rewrite(f_out);
    with grid do begin
      for r := 0 to rowcount-1 do begin
        for c := 0 to colcount-2 do
          write(f_out,Cells[c,r],';');
        writeln(f_out,Cells[colcount-1,r]);
      end;
    end;
  Finally
    Closefile(f_out);
  end;
end;

//----------------------------------------------------------------------------

procedure gridExchangeRows(var grid: TStringGrid; i, j: Integer );
Var
  k: Integer;
begin
  for k:= 0 to grid.ColCount-1 do grid.Cols[k].Exchange(i,j)
end;

//----------------------------------------------------------------------------

procedure gridSort(var grid: TStringGrid; byColumn: Integer; iLo, iHi: Integer; CaseSensitive: Boolean);
var
  Lo, Hi: Integer;
  P: String;
begin
  repeat
    Lo := iLo;
    Hi := iHi;
    P := grid.Cells[byColumn, (iLo + iHi) shr 1];
    repeat
      case CaseSensitive of
        False:
          begin
            while CompareText(grid.Cells[byColumn, Lo], P) < 0 do Inc(Lo);
            while CompareText(grid.Cells[byColumn, Hi], P) > 0 do Dec(Hi);
          end;
        true:
          begin
            while CompareStr(grid.Cells[byColumn, Lo], P) < 0 do Inc(Lo);
            while CompareStr(grid.Cells[byColumn, Hi], P) > 0 do Dec(Hi);
          end;
      end;    {case}
      if Lo <= Hi then begin
        If Lo <> Hi then gridExchangeRows(grid, Lo, Hi);
        Inc(Lo);
        Dec(Hi);
      end;
    until Lo > Hi;
    if iLo < Hi then gridSort(grid, byColumn, iLo, Hi, CaseSensitive);
    iLo := Lo;
  until Lo >= iHi;
end;

//----------------------------------------------------------------------------

procedure gridInvert(var grid: TStringGrid);
Var
  i, j: Integer;
begin
  with grid do begin
    i:= Fixedrows;
    j:= Rowcount-2;
    while i < j do begin
      gridExchangeRows(grid, I, J);
      Inc( i );
      Dec( j );
    end; { While }
  end;
end;

//----------------------------------------------------------------------------

procedure memoLoad(var memo: TMemo; fichier: string);
begin
  if FileExistsUTF8(fichier) { *Converted from FileExists* } then
    memo.Lines.LoadFromFile(fichier);
end;

//----------------------------------------------------------------------------

procedure memoSave(memo: TMemo; fichier: string);
begin
  memo.Lines.SaveToFile(fichier);
end;

//----------------------------------------------------------------------------

procedure listboxLoad(var liste: TListBox; fichier: string);
begin
  if FileExistsUTF8(fichier) { *Converted from FileExists* } then
    liste.Items.LoadFromFile(fichier);
end;

//----------------------------------------------------------------------------

procedure listboxSave(liste: TListBox; fichier: string);
begin
  liste.Items.SaveToFile(fichier);
end;

//----------------------------------------------------------------------------

function TryStrToInt(const S: string; out Value: Integer): Boolean;
var
  E: Integer;
begin
  Val(S, Value, E);
  Result := E = 0;
end;

//----------------------------------------------------------------------------

function AnsiReplaceText(str, sub1, sub2: String): String;
var
  aPos: Integer;
  rslt: String;
begin
  aPos := Pos(sub1, str);
  rslt := '';
  while (aPos <> 0) do begin
    rslt := rslt + Copy(str, 1, aPos - 1) + sub2;
    Delete(str, 1, aPos + Length(sub1)-1);
    aPos := Pos(sub1, str);
  end;
  Result := rslt + str;
end;

//----------------------------------------------------------------------------

function replaceString(str, sub1, sub2: string): string;
var
  n, l : integer;
begin
  n := Pos( sub1, str );
  l := Length( sub1 );
  while(n > 0)do begin
    Delete( str, n, l );
    Insert( sub2, str, n );
    n := Pos( sub1, str );
  end;
  Result := str;
end;

//----------------------------------------------------------------------------

procedure copyFile(old_file, new_file: string);
var
  FromF, ToF: file;
  NumRead, NumWritten: Integer;
  Buf: array[1..2048] of Char;
begin
  AssignFile(FromF, old_file);
  Reset(FromF, 1);
  AssignFile(ToF, new_file);
  Rewrite(ToF, 1);
  repeat
    BlockRead(FromF, Buf, SizeOf(Buf), NumRead);
    BlockWrite(ToF, Buf, NumRead, NumWritten);
  until (NumRead = 0) or (NumWritten <> NumRead);
  CloseFile(FromF);
  CloseFile(ToF);
end;

//----------------------------------------------------------------------------

procedure textReplace(old_file, new_file, old, new: string);
var
  s: string;
  f_in, f_tmp: text;
begin
  {$I-}
  s := '';
  AssignFile(f_in, old_file);
  Reset(f_in);
  if IOResult = 0 then begin
    AssignFile(f_tmp, 'temp.$$$');
    Rewrite(f_tmp);
    if IOResult = 0 then begin
      while not EOF(f_in) do begin
        Readln(f_in, s);
        Writeln(f_tmp, replaceString(s, old, new));
      end;
      CloseFile(f_tmp);
    end;
    CloseFile(f_in);
  end;
  {$I+}
  copyFile('temp.$$$', new_file);
  Erase(f_tmp);
end;

//----------------------------------------------------------------------------

function IncludeTrailingPathDelimiter(const S: string): string;
begin
  if s[length(S)] = '\' then
    Result := S
  else
    Result := S + '\';
end;

//----------------------------------------------------------------------------

function HexString(const t: PByteArray; size: Integer): string;
var
  i: Integer;
begin
  Result := '';
  for i := 0 to size - 1 do Result := Result + IntToHex(t^[i], 2);
end;

//----------------------------------------------------------------------------

function HexDump(const t: PByteArray; size: Integer): string;
var
  i: integer;
  j: integer;
  s: string;
begin
  if size = 0 then exit;
  s := '';
  Result := 'Addr :  0  1  2  3  4  5  6  7   8  9  A  B  C  D  E  F' + CrLf;
  for i := 0 to size - 1 do begin
    if (i mod 16) = 0 then Result := Result + IntToHex(i, 4) + ' : ';
    j := t^[i];
    Result := Result + IntToHex(j, 2) + ' ';
    if j > 31 then s := s + Chr(j) else s := s + '.';
    if (i mod 16) = 15 then begin
      Result := Result + ': ' + s + CrLf;
      s := '';
    end else if (i mod 8) = 7 then begin
      Result := Result + ' ';
      s := s + ' ';
    end;
  end;
  if ((size-1) mod 16) < 15 then begin
    for i := ((size-1) mod 16) to 14 do Result := Result + '   ';
    if ((size-1) mod 16) < 7 then Result := Result + ' ';
    Result := Result + ': ' + s + CrLf;
  end;
end;

//----------------------------------------------------------------------------

function readCSV(filename: string; var MyArray : TStringMatrix; var fields: TStringList): longint;
var
  F: TextFile;
  r: LongInt;
  rec: string;
  Elements: TStringList;
  n, nbCols, c: byte;
begin
  r := 0;
  fields.Clear;
  setLength(MyArray,0);
  try
    Assignfile(F,filename);
    Reset(F);
    Elements := TStringList.Create;
    if Eof(F) then raise EInvalidFile.Create('Invalid file '+filename+'. Program cannot read data');
    Readln(F,rec);
    nbCols := splitString(rec, ';', fields);
    while not Eof(f) do begin
      Readln(F,rec);
      Inc(r);
      n := splitString(rec, ';', Elements);
      if n > nbCols then
        raise EInvalidFile.Create('Invalid file'+filename+'. Too much columns in line '+IntToStr(r));
      if n < nbCols then
        raise EInvalidFile.Create('Invalid file'+filename+'. Not enough columns in line '+IntToStr(r));
      setLength(MyArray,r);
      setLength(MyArray[r-1],nbCols);
      for c := 0 to n-1 do MyArray[r-1,c] := Elements[c];
    end;
  Finally
    Closefile(F);
    Elements.Free;
    result := r;
  end;
end;

//----------------------------------------------------------------------------

procedure saveCSV(filename: string; var MyArray : TStringMatrix; var fields: TStringList);
var
  F: TextFile;
  i: integer;
  j: byte;
  Elements: TStringList;
begin
  if FileExists(FileName) then begin
    if FileExists(FileName+'S') then DeleteFile(FileName+'S');
    if not RenameFile(FileName, filename+'S') then begin
      warnErr('Erreur de renommage de fichier');
      exit;
    end;
  end;
  AssignFile(F, FileName);
  Rewrite(F);
  Elements := TStringList.Create;
  Writeln(F, joinStrings(fields, ';'));
  for i := 0 to length(MyArray)-1 do begin
    Elements.Clear;
    for j := 0 to length(MyArray[i])-1 do Elements.Add(MyArray[i][j]);
    Writeln(F, joinStrings(Elements, ';'));
  end;
  CloseFile(F);
end;

//----------------------------------------------------------------------------

procedure readDescription(FileName: string; var descList: TDescList);
var
  fields: TStringList;
  MyArray: TStringMatrix;
  n, i: integer;
begin
  descList.Clear;
  if not FileExists(FileName) then exit;
  fields := TStringList.Create;
  n := readCSV(filename, MyArray, fields);
  if fields.Count > 2 then begin
    warnErr('Le fichier '+Filename+' a plus de 2 champs!');
    exit;
  end;
  if fields.Count < 2 then begin
    warnErr('Le fichier '+Filename+' a moins de 2 champs!');
    exit;
  end;
  fields.Free;
  for i := 0 to n-1 do
    descList.AddDesc(
      StrToInt(MyArray[i][0]),
      unQuote(MyArray[i][1])
      );
  for i := n-1 downto 0 do MyArray[i] := nil;
  Finalize(MyArray);
end;

//----------------------------------------------------------------------------

function findDescription(descList: TDescList; id: integer): string;
var
  i: integer;
begin
  for i := 0 to descList.Count-1 do begin
    if descList.Descriptions[i]^.FId = id then begin
      Result := descList.Descriptions[i]^.FName;
      exit;
    end;
  end;
  Result := 'Id '+IntToStr(id)+' inconnu!';
end;

end.
